// Code generated by smithy-go-codegen DO NOT EDIT.

// Package sso provides the API client, operations, and parameter types for AWS
// Single Sign-On.
//
// AWS Single Sign-On Portal is a web service that makes it easy for you to assign
// user access to AWS SSO resources such as the user portal. Users can get AWS
// account applications and roles assigned to them and get federated into the
// application. For general information about AWS SSO, see What is AWS Single
// Sign-On?
// (https://docs.aws.amazon.com/singlesignon/latest/userguide/what-is.html) in the
// AWS SSO User Guide. This API reference guide describes the AWS SSO Portal
// operations that you can call programatically and includes detailed information
// on data types and errors. AWS provides SDKs that consist of libraries and sample
// code for various programming languages and platforms, such as Java, Ruby, .Net,
// iOS, or Android. The SDKs provide a convenient way to create programmatic access
// to AWS SSO and other AWS services. For more information about the AWS SDKs,
// including how to download and install them, see Tools for Amazon Web Services
// (http://aws.amazon.com/tools/).
package sso
